/* NAT for netfilter; shared with compatibility layer. */

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netfilter_ipv4.h>
#include <net/checksum.h>
#include <net/icmp.h>
#include <net/ip.h>
#include <net/tcp.h>  /* For tcp_prot in getorigdst */
#include "NAT/ip_nat.h"
#include "NAT/ip_nat_protocol.h"
#include "NAT/ip_nat_core.h"
#include "NAT/ip_nat_helper.h"
#include "common/listhelp.h"

size_t ip_nat_offset;

#define IP_NAT_HTABLE_SIZE 64

static struct list_head bysource[IP_NAT_HTABLE_SIZE];
static struct list_head byipsproto[IP_NAT_HTABLE_SIZE];
LIST_HEAD(protos);
static LIST_HEAD(helpers);

extern struct ip_nat_protocol unknown_nat_protocol;

/* We keep extra hashes for each conntrack, for fast searching. */
static inline size_t 
hash_by_ipsproto(u_int32_t src, u_int32_t dst, u_int16_t proto)
{
	/* Modified src and dst, to ensure we don't create two
           identical streams. */
	return (src + dst + proto) % IP_NAT_HTABLE_SIZE;
}

static inline size_t 
hash_by_src(const struct ip_conntrack_manip *manip, u_int16_t proto)
{
	/* Original src, to ensure we map it consistently if poss. */
	return (manip->ip + manip->u.all + proto) % IP_NAT_HTABLE_SIZE;
}

static void ip_nat_cleanup_conntrack(struct ip_conntrack *conn)
{
	struct ip_nat_info *info = (void *)conn + ip_nat_offset;

	if (info->bysource.conntrack) {
		LIST_DELETE(&bysource[hash_by_src(&conn->tuplehash[0]
						  .tuple.src,
						  conn->tuplehash[0]
						  .tuple.dst.protonum)],
			    &info->bysource);
	}

	if (info->byipsproto.conntrack) {
		LIST_DELETE(&byipsproto[hash_by_ipsproto(conn->tuplehash[1]
							 .tuple.src.ip,
							 conn->tuplehash[1]
							 .tuple.dst.ip,
							 conn->tuplehash[1]
							 .tuple.dst.protonum)],
			    &info->byipsproto);
	}

#ifdef DEBUG_CONNTRACK
	/* Look to see if there's an expect in this deleted conntrack;
	   catches a tricky bug in helpers. */
	{
		struct list_head *i;

		for (i = expect_list.next; i != &expect_list; i = i->next) 
			if ((void *)i >= (void *)conn
			    && (void *)i < (void *)(conn+1))
				DEBUGP("Deleting conn %p: live expect!\n", i);
	}
#endif /*DEBUG_CONNTRACK*/
}

/* We do checksum mangling, so if they were wrong before they're still
 * wrong.  Also works for incomplete packets (eg. ICMP dest
 * unreachables.) */
u_int16_t 
ip_nat_cheat_check(u_int32_t oldvalinv, u_int32_t newval, u_int16_t oldcheck)
{
	u_int32_t diffs[] = { oldvalinv, newval };
	return csum_fold(csum_partial((char *)diffs, sizeof(diffs),
				      oldcheck^0xFFFF));
}

static inline int cmp_proto(const struct ip_nat_protocol *i, int proto)
{
	return i->protonum == proto;
}

struct ip_nat_protocol *
find_nat_proto(u_int16_t protonum)
{
	struct ip_nat_protocol *i = LIST_FIND(&protos, cmp_proto, 
					      struct ip_nat_protocol *,
					      protonum);
	if (!i)
		i = &unknown_nat_protocol;
	return i;
}

/* Is this tuple already taken? (not by us)*/
int
ip_nat_used_tuple(const struct ip_conntrack_tuple *tuple,
		  const struct ip_conntrack *ignored_conntrack)
{
	/* Conntrack tracking doesn't keep track of outgoing tuples; only
	   incoming ones.  NAT means they don't have a fixed mapping,
	   so we invert the tuple and look for the incoming reply.

	   We could keep a separate hash if this proves too slow. */
	struct ip_conntrack_tuple reply;

	invert_tuplepr(&reply, tuple);
	return (int)ip_conntrack_find(&reply, ignored_conntrack);
}

/* Does tuple + the source manip come within the range mr */
static int
in_range(const struct ip_conntrack_tuple *tuple,
	 const struct ip_conntrack_manip *manip,
	 const struct ip_nat_multi_range *mr)
{
	struct ip_nat_protocol *proto = find_nat_proto(tuple->dst.protonum);
	unsigned int i;
	struct ip_conntrack_tuple newtuple = { *manip, tuple->dst };

	/* For the moment, this is true. --RR */
	IP_CONNTRACK_ASSERT(mr->rangesize == 1);
	for (i = 0; i < mr->rangesize; i++) {
		/* If we are allowed to map IPs, then we must be in the
		   range specified, otherwise we must be unchanged. */
		if (mr->range[i].flags & IP_NAT_RANGE_MAP_IPS) {
			if (ntohl(newtuple.src.ip) < ntohl(mr->range[i].min_ip)
			    || (ntohl(newtuple.src.ip)
				> ntohl(mr->range[i].max_ip)))
				continue;
		} else {
			if (newtuple.src.ip != tuple->src.ip)
				continue;
		}

		if ((mr->range[i].flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		    && proto->in_range(&newtuple, IP_NAT_MANIP_SRC,
				       &mr->range[i].min, &mr->range[i].max))
			return 1;
	}
	return 0;
}

static inline int
src_cmp(const struct ip_nat_hash *i,
	const struct ip_conntrack_tuple *tuple,
	const struct ip_nat_multi_range *mr)	
{
	return (i->conntrack->tuplehash[0].tuple.dst.protonum
		== tuple->dst.protonum
		&& i->conntrack->tuplehash[0].tuple.src.ip
		== tuple->src.ip
		&& i->conntrack->tuplehash[0].tuple.src.u.all
		== tuple->src.u.all
		&& in_range(tuple, &i->conntrack->tuplehash[0].tuple.src, mr));
}

/* Only called for SRC manip */
static struct ip_conntrack_manip *
find_appropriate_src(const struct ip_conntrack_tuple *tuple,
		     const struct ip_nat_multi_range *mr)
{
	unsigned int h = hash_by_src(&tuple->src, tuple->dst.protonum);
	struct ip_nat_hash *i;

	i = LIST_FIND(&bysource[h], src_cmp, struct ip_nat_hash *, tuple, mr);
	if (i)
		return &i->conntrack->tuplehash[0].tuple.src;
	else
		return NULL;
}

/* If it's really a local destination manip, it may need to do a
   source manip too. */
static int
do_extra_mangle(u_int32_t var_ip, u_int32_t *other_ipp)
{
	struct rtable *rt;

	/* FIXME: IPTOS_TOS(iph->tos) --RR */
	if (ip_route_output(&rt, var_ip, 0, 0, 0) != 0) {
		DEBUGP("do_extra_mangle: Can't get route to %u.%u.%u.%u\n",
		       IP_PARTS(var_ip));
		return 0;
	}

	*other_ipp = rt->rt_src;
	ip_rt_put(rt);
	return 1;
}

/* Simple way to iterate through all. */
static inline int fake_cmp(const struct ip_nat_hash *i,
			   u_int32_t src, u_int32_t dst, u_int16_t protonum,
			   unsigned int *score)
{
	/* Compare backwards: we're dealing with OUTGOING tuples, and
           inside the conntrack is the REPLY tuple. */
	if (i->conntrack->tuplehash[1].tuple.src.ip == dst
	    && i->conntrack->tuplehash[1].tuple.dst.ip == src
	    && i->conntrack->tuplehash[1].tuple.dst.protonum == protonum)
		(*score)++;
	return 0;
}

static inline unsigned int
count_maps(u_int32_t src, u_int32_t dst, u_int16_t protonum)
{
	unsigned int score = 0;

	LIST_FIND(&byipsproto[hash_by_ipsproto(src, dst, protonum)],
		  fake_cmp, struct ip_nat_hash *, src, dst, protonum, &score);

	return score;
}

/* For [FUTURE] fragmentation handling, we want the least-used
   src-ip/dst-ip/proto triple.  Fairness doesn't come into it.  Thus
   if the range specifies 1.2.3.4 ports 10000-10005 and 1.2.3.5 ports
   1-65535, we don't do pro-rata allocation based on ports; we choose
   the ip with the lowest src-ip/dst-ip/proto usage.

   If an allocation then fails (eg. all 6 ports used in the 1.2.3.4
   range), we eliminate that and try again.  This is not the most
   efficient approach, but if you're worried about that, don't hand us
   ranges you don't really have.  */
static struct ip_nat_range *
find_best_ips_proto(struct ip_conntrack_tuple *tuple,
		    const struct ip_nat_multi_range *mr,
		    enum ip_nat_manip_type maniptype,
		    const struct net_device *device)
{
	unsigned int i;
	struct {
		const struct ip_nat_range *range;
		unsigned int score;
		struct ip_conntrack_tuple tuple;
	} best = { NULL,  0xFFFFFFFF };
	u_int32_t *var_ipp, *other_ipp, saved_ip;
#ifdef DEBUG_CONNTRACK
	unsigned int loopmax = 0;
#endif

	if (maniptype == IP_NAT_MANIP_SRC) {
		var_ipp = &tuple->src.ip;
		saved_ip = tuple->dst.ip;
		other_ipp = &tuple->dst.ip;
	} else {
		var_ipp = &tuple->dst.ip;
		saved_ip = tuple->src.ip;
		other_ipp = &tuple->src.ip;
	}

	IP_CONNTRACK_ASSERT(mr->rangesize >= 1);
	for (i = 0; i < mr->rangesize; i++) {
		u_int32_t minip, maxip;

		/* Don't do ranges which are already eliminated. */
		if (mr->range[i].flags & IP_NAT_RANGE_FULL) continue;

		if (mr->range[i].flags & IP_NAT_RANGE_MAP_IPS) {
			minip = mr->range[i].min_ip;
			maxip = mr->range[i].max_ip;
		} else
			minip = maxip = *var_ipp;

		for (*var_ipp = minip;
		     ntohl(*var_ipp) <= ntohl(maxip);
		     *var_ipp = htonl(ntohl(*var_ipp) + 1)) {
			unsigned int score;
			IP_CONNTRACK_ASSERT(++loopmax < 1000);

			/* Reset the other ip in case it was mangled by
			 * do_extra_mangle last time. */
			*other_ipp = saved_ip;

			if (!device
			    && maniptype == IP_NAT_MANIP_DST
			    && !do_extra_mangle(*var_ipp, other_ipp)) {
				/* Can't route?  This whole range part is
				 * probably screwed, but keep trying
				 * anyway. */
				continue;
			}

			/* Count how many others map onto this. */
			score = count_maps(tuple->src.ip, tuple->dst.ip,
					   tuple->dst.protonum);
			if (score < best.score) {
				/* Optimization: doesn't get any better than
				   this. */
				if (score == 0)
					return (struct ip_nat_range *)
						&mr->range[i];

				best.score = score;
				best.tuple = *tuple;
				best.range = &mr->range[i];
			}
		}
	}
	*tuple = best.tuple;

	/* Discard const. */
	return (struct ip_nat_range *)best.range;
}

static int
get_unique_tuple(struct ip_conntrack_tuple *tuple,
		 const struct ip_conntrack_tuple *orig_tuple,
		 enum ip_nat_manip_type maniptype,
		 struct ip_nat_multi_range *mr,
		 const struct ip_conntrack *conntrack,
		 const struct net_device *device)
{
	struct ip_nat_protocol *proto
		= find_nat_proto(orig_tuple->dst.protonum);
	struct ip_nat_range *rptr;
	unsigned int i;
	int ret;

#ifdef DEBUG_CONNTRACK
	unsigned int loopmax = 0;
#endif

	/* 1) If this srcip/proto/src-proto-part is currently mapped,
	   and that same mapping gives a unique tuple within the given
	   range, use that.

	   This is only required for source (ie. NAT/masq) mappings.
	   So far, we don't do local source mappings, so multiple
	   manips not an issue.  */
	if (maniptype == IP_NAT_MANIP_SRC) {
		struct ip_conntrack_manip *manip;

		manip = find_appropriate_src(orig_tuple, mr);
		if (manip) {
			/* Apply same source manipulation. */
			*tuple = ((struct ip_conntrack_tuple)
				  { *manip, orig_tuple->dst });
			DEBUGP("get_unique_tuple: Found current src mapping\n");
			return 1;
		}
	}

	/* 2) Select the least-used IP/proto combination in the given
	   range.
	*/
	*tuple = *orig_tuple;
	while ((rptr = find_best_ips_proto(tuple, mr, maniptype, device))
	       != NULL) {
		IP_CONNTRACK_ASSERT(++loopmax < 1000);

		/* 3) The per-protocol part of the manip is made to
		   map into the range to make a unique tuple. */

		/* Only bother mapping if it's not already in range
		   and unique */
		if ((!(rptr->flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		     || proto->in_range(tuple, maniptype,
					&rptr->min, &rptr->max))
		    && !ip_nat_used_tuple(tuple, conntrack)) {
			ret = 1;
			goto clear_fulls;
		}
		else if (proto->unique_tuple(tuple, rptr, maniptype, 
					     conntrack)) {
			/* Must be unique. */
			IP_CONNTRACK_ASSERT(!ip_nat_used_tuple(tuple,
							       conntrack));
			ret = 1;
			goto clear_fulls;
		}

		/* Eliminate that from range, and try again. */
		rptr->flags |= IP_NAT_RANGE_FULL;
		*tuple = *orig_tuple;
	}

	ret = 0;

 clear_fulls:
	/* Clear full flags. */
	IP_CONNTRACK_ASSERT(mr->rangesize >= 1);
	for (i = 0; i < mr->rangesize; i++)
		mr->range[i].flags &= ~IP_NAT_RANGE_FULL;

	return ret;
}

static inline int 
helper_cmp(struct ip_nat_helper *helper, 
	   u_int16_t protocol, 
	   u_int16_t protocol_dst)
{
	return (protocol == helper->protocol
		&& protocol_dst == helper->protocol_dst);
}

/* Set up which hooks manips are to be done on: index by non-local/maniptype */
/* Packet comes in off network, user asks for a source manip, so we
   delay.  We need to catch it in POST_ROUTING: if we see it at
   LOCAL_IN, then we undo our conntrack changes (we don't alter source
   address on LOCAL_IN, because lower down code uses route to create
   replies, ie. TCP in ip_send_reply. */
static unsigned int orig_hooks[2][2] 
= { { NF_IP_POST_ROUTING,		/* !local, src manip */
      NF_IP_PRE_ROUTING },		/* !local, dst manip */
    { NF_IP_POST_ROUTING,	 	/* local, src manip */
      NF_IP_LOCAL_OUT }			/* local, dst manip */
};
/* Where to manip the reply packets (will be reverse manip). */
static unsigned int reply_hooks[2][2] 
= { { NF_IP_PRE_ROUTING,		/* !local, src manip */
      NF_IP_POST_ROUTING },		/* !local, dst manip */
    { NF_IP_PRE_ROUTING,		/* local, src manip */
      NF_IP_POST_ROUTING }		/* local, dst manip */
};

unsigned int
ip_nat_setup_info(struct ip_conntrack *conntrack,
		  enum ip_nat_manip_type maniptype,
		  struct ip_nat_multi_range *mr,
		  const struct net_device *device)
{
	struct ip_conntrack_tuple new_tuple, inv_tuple, reply;
	struct ip_nat_manipulation *m, *m_rev;
	struct ip_nat_info *info = (void *)conntrack + ip_nat_offset;
	const struct ip_conntrack_tuple *orig_tuple 
		= &conntrack->tuplehash[0].tuple;

	info->dirinfo[0].hook = orig_hooks[!device][maniptype];
	info->dirinfo[1].hook = reply_hooks[!device][maniptype];

	m = info->dirinfo[0].manipulation;
	m_rev = info->dirinfo[1].manipulation;

	if (!get_unique_tuple(&new_tuple, orig_tuple, maniptype, mr, conntrack,
			      device)) {
		DEBUGP("ip_nat_alloc_binding: Can't get unique tuple.\n");
		return NF_DROP;
	}

	/* We now have two tuples (SRCIP/SRCPT/DSTIP/DSTPT):
	   the original (A/B/C/D') and the mangled one (E/F/G/H').

	   We're only allowed to work with the SRC per-proto part,
	   so we create inverses of both to start, then derive the other
	   fields we need.  */

	/* Reply connection: simply invert the new tuple (G/H/E/F') */
	invert_tuplepr(&reply, &new_tuple);

	/* Alter conntrack table so it recognizes replies. */
	ip_conntrack_alter_reply(conntrack, &reply);

	/* Create inverse of original: C/D/A/B' */
	invert_tuplepr(&inv_tuple, orig_tuple);

	/* Has source changed?. */
	if (memcmp(&new_tuple.src, &orig_tuple->src, sizeof(new_tuple.src))
	    != 0) {
		/* In this direction, a source manip. */
		m->maniptype = IP_NAT_MANIP_SRC;
		m->manip = new_tuple.src;
		m++;
		info->dirinfo[0].num_manips++;

		/* In the reverse direction, a destination manip. */
		m_rev->maniptype = IP_NAT_MANIP_DST;
		m_rev->manip = orig_tuple->src;
		m_rev++;
		info->dirinfo[1].num_manips++;
	}

	/* Has destination changed? */
	if (memcmp(&new_tuple.dst, &orig_tuple->dst, sizeof(new_tuple.dst))
	    != 0) {
		/* In this direction, a destination manip */
		m->maniptype = IP_NAT_MANIP_DST;
		m->manip = reply.src;
		m++;
		info->dirinfo[0].num_manips++;

		/* In the reverse direction, a source manip. */
		m_rev->maniptype = IP_NAT_MANIP_SRC;
		m_rev->manip = inv_tuple.src;
		m_rev++;
		info->dirinfo[1].num_manips++;
	}

	/* If there's a helper, assign it; based on new tuple/ */
	info->helper = LIST_FIND(&helpers, helper_cmp, struct ip_nat_helper *,
				 new_tuple.dst.protonum,
				 new_tuple.dst.u.all);
	return NF_ACCEPT;
}

void place_in_hashes(struct ip_conntrack *conntrack)
{
	struct ip_nat_info *info = (void *)conntrack + ip_nat_offset;

	unsigned int hash
		= hash_by_src(&conntrack->tuplehash[0].tuple.src,
			      conntrack->tuplehash[0].tuple.dst.protonum);
	/* Put it in bysource hash. */
	IP_CONNTRACK_ASSERT(!info->bysource.conntrack);
	list_prepend(&bysource[hash], &info->bysource);
	info->bysource.conntrack = conntrack;

	/* We place packet as seen OUTGOUNG in byips_proto hash
           (ie. reverse dst and src of reply packet. */
	hash = hash_by_ipsproto(conntrack->tuplehash[1].tuple.dst.ip,
				conntrack->tuplehash[1].tuple.src.ip,
				conntrack->tuplehash[1].tuple.dst.protonum);
	info->byipsproto.conntrack = conntrack;
	list_prepend(&byipsproto[hash], &info->byipsproto);
}

static void
manip_pkt(u_int16_t proto, struct iphdr *iph, size_t len,
	  const struct ip_conntrack_manip *manip,
	  enum ip_nat_manip_type maniptype)
{
	find_nat_proto(proto)->manip_pkt(iph, len, manip, maniptype);

	if (maniptype == IP_NAT_MANIP_SRC) {
		iph->check = ip_nat_cheat_check(~iph->saddr, manip->ip,
						iph->check);
		iph->saddr = manip->ip;
	} else {
		iph->check = ip_nat_cheat_check(~iph->daddr, manip->ip,
						iph->check);
		iph->daddr = manip->ip;
	}
#if 0
	if (ip_fast_csum((u8 *)iph, iph->ihl) != 0)
		DEBUGP("IP: checksum on packet bad.\n");

	if (proto == IPPROTO_TCP) {
		void *th = (u_int32_t *)iph + iph->ihl;
		if (tcp_v4_check(th, len - 4*iph->ihl, iph->saddr, iph->daddr,
				 csum_partial((char *)th, len-4*iph->ihl, 0)))
			DEBUGP("TCP: checksum on packet bad\n");
	}
#endif
}

/* Do packet manipulations according to binding. */
unsigned int
do_binding(struct ip_conntrack_tuple_hash *h,
	   enum ip_conntrack_info conntrackinfo,
	   struct sk_buff **pskb)
{
	unsigned int i;
	struct ip_nat_info *info = (void *)h->ctrack + ip_nat_offset;
	int dir = DIRECTION(h);

	IP_CONNTRACK_ASSERT(info->dirinfo[dir].num_manips < 3);
	for (i = 0; i < info->dirinfo[dir].num_manips; i++) {
		manip_pkt((*pskb)->nh.iph->protocol, (*pskb)->nh.iph, 
			  (*pskb)->len,
			  &info->dirinfo[dir].manipulation[i].manip,
			  info->dirinfo[dir].manipulation[i].maniptype);
	}

	if (info->helper)
		return info->helper->help(h, info, conntrackinfo, pskb);
	else return NF_ACCEPT;
}

void 
icmp_reply_translation(struct sk_buff *skb, 
		       struct ip_conntrack *conntrack,
		       int conn_dir)
{
	struct iphdr *iph = skb->nh.iph;
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct iphdr *inner = (struct iphdr *)(hdr + 1);
	struct ip_nat_info *info = (void *)conntrack + ip_nat_offset;
	size_t datalen = skb->len - ((void *)inner - (void *)iph);
	unsigned int i;

	/* Map inner packet first; do this REVERSE of normal. */
	IP_CONNTRACK_ASSERT(info->dirinfo[conn_dir].num_manips < 3);
	for (i = 0; i < info->dirinfo[conn_dir].num_manips; i++) {
		manip_pkt(inner->protocol, inner, 
			  skb->len - ((void *)inner - (void *)iph),
			  &info->dirinfo[conn_dir].manipulation[i].manip,
			  !info->dirinfo[conn_dir].manipulation[i].maniptype);
	}

	/* Now outer packet. */
	/* Note: May not be from a NAT'd host, but probably safest to
	   do translation always as if it came from the host itself
	   (even though a "host unreachable" coming from the host
	   itself is a bit wierd).

	   More explanation: some people use NAT for anonomising.
	   Also, CERT recommends dropping all packets from private IP
	   addresses (although ICMP errors from internal links with
	   such addresses are not too uncommon, as Alan Cox points
	   out) */

	/* Use mapping to map outer packet: 0 give no per-proto mapping */
	IP_CONNTRACK_ASSERT(info->dirinfo[conn_dir].num_manips < 3);
	for (i = 0; i < info->dirinfo[conn_dir].num_manips; i++) {
		manip_pkt(0, iph, skb->len,
			  &info->dirinfo[conn_dir].manipulation[i].manip,
			  info->dirinfo[conn_dir].manipulation[i].maniptype);
	}

	/* Since we mangled inside ICMP packet, recalculate its
	   checksum from scratch.  (Hence the handling of incorrect
	   checksums above, so we don't accidentally fix one.) */
	hdr->checksum = 0;
	hdr->checksum = ip_compute_csum((unsigned char *)hdr,
					sizeof(*hdr) + datalen);
}

/* We've already marked the conntrack for NAT, but now the packet is
   coming in to the local box, it's simplest to undo it. */
void ip_nat_undo(struct ip_conntrack *conntrack)
{
	struct ip_conntrack_tuple tuple;

	invert_tuplepr(&tuple, &conntrack->tuplehash[0].tuple);
	ip_conntrack_alter_reply(conntrack, &tuple);
}

int ip_nat_helper_register(struct ip_nat_helper *me)
{
	int ret = 0;
	unsigned long flags;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	if (LIST_FIND(&helpers, helper_cmp, struct ip_nat_helper *,
		      me->protocol, me->protocol_dst))
		ret = -EBUSY;
	else {
		list_prepend(&helpers, me);
		MOD_INC_USE_COUNT;
	}
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);

	return ret;
}

void ip_nat_helper_unregister(struct ip_nat_helper *me)
{
	unsigned long flags;
	unsigned int i;
	MOD_DEC_USE_COUNT;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	/* Find anything using it, and umm, kill them.  We can't turn
	   them into normal connections: if we've adjusted SYNs, then
	   they'll ackstorm.  So we just drop it.  We used to just
	   bump module count when a connection existed, but that
	   forces admins to gen fake RSTs or bounce box, either of
	   which is just a long-winded way of making things
	   worse. --RR */
	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
		/* Subtle: delete while traversing. */
		struct list_head *prev;

		for (prev = &ip_conntrack_hash[i];
		     prev->next != &ip_conntrack_hash[i];
		     prev = prev->next) {
			struct ip_conntrack_tuple_hash *this
				= (void *)prev->next;
			struct ip_nat_info *info 
				= (void *)this->ctrack + ip_nat_offset;
			if (info->helper == me)
				ip_conntrack_delete(this->ctrack);
		}
	}
	LIST_DELETE(&helpers, me);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
}

#ifdef HAVE_NF_GETSOCKOPT
/* Fast function for those who don't want to parse /proc (and I don't
   blame them. */
/* Reversing the socket's dst/src point of view gives us the reply
   mapping. */
static int 
getorigdst(struct sock *sk, int optval, void *user, int *len)
{
	unsigned long flags;
	struct ip_conntrack_tuple_hash *h;
	struct ip_conntrack_tuple tuple = { { sk->rcv_saddr, { sk->sport },
					      0 },
					    { sk->daddr, { sk->dport },
					      IPPROTO_TCP } };

	/* We only do TCP at the moment: is there a better way? */
	if (strcmp(sk->prot->name, "TCP") != 0) {
		DEBUGP("SO_ORIGINAL_DST: Not a TCP socket\n");
		return -ENOPROTOOPT;
	}

	if (*len != sizeof(struct sockaddr_in)) {
		DEBUGP("SO_ORIGINAL_DST: len %u not %u\n", 
		       *len, sizeof(struct sockaddr_in));
		return -EINVAL;
	}

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	h = ip_conntrack_find(&tuple, NULL);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	if (h) {
		struct sockaddr_in sin 
			= { AF_INET,
			    h->ctrack->tuplehash[0].tuple.dst.u.tcp.port,
			    { h->ctrack->tuplehash[0].tuple.dst.ip } };
		if (copy_to_user(user, &sin, sizeof(sin)) != 0)
			return -EFAULT;
		else
			return 0;
	}

	DEBUGP("SO_ORIGINAL_DST: Can't find %u.%u.%u.%u/%u-%u.%u.%u.%u/%u.\n",
	       IP_PARTS(tuple.src.ip), ntohs(tuple.src.u.tcp.port),
	       IP_PARTS(tuple.dst.ip), ntohs(tuple.dst.u.tcp.port));
	return -ENOPROTOOPT;
}

static struct nf_sockopt_ops so_getorigdst
= { { NULL, NULL }, PF_INET, 
    0, 0, NULL, /* Setsockopts */
    SO_ORIGINAL_DST, SO_ORIGINAL_DST+1, &getorigdst };
#endif /* HAVE_NF_GETSOCKOPT */
int __init ip_nat_init(void)
{
	size_t i;

	/* Reserve space in conntrack structure. */
	ip_nat_offset = struct_reserve(ip_conntrack, 
				       sizeof(struct ip_nat_info));

	/* Sew in builtin protocols. */
	list_append(&protos, &ip_nat_protocol_tcp);
	list_append(&protos, &ip_nat_protocol_udp);
	list_append(&protos, &ip_nat_protocol_icmp);

	for (i = 0; i < IP_NAT_HTABLE_SIZE; i++) {
		INIT_LIST_HEAD(&bysource[i]);
		INIT_LIST_HEAD(&byipsproto[i]);
	}

	IP_CONNTRACK_ASSERT(ip_conntrack_destroyed == NULL);
	ip_conntrack_destroyed = &ip_nat_cleanup_conntrack;

#ifdef HAVE_NF_GETSOCKOPT
	return nf_register_sockopt(&so_getorigdst);
#else
	return 0;
#endif
}

void ip_nat_cleanup(void)
{
	/* Unreserve space in conntrack structure. */
	struct_unreserve(ip_conntrack, ip_nat_offset,
			 sizeof(struct ip_nat_info));

	ip_conntrack_destroyed = NULL;
#ifdef HAVE_NF_GETSOCKOPT
	nf_unregister_sockopt(&so_getorigdst);
#endif
}
